import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit


@jit(nopython=True)
def alifestd_mark_num_descendants_asexual_fast_path(
    ancestor_ids: np.ndarray,
) -> np.ndarray:
    """Implementation detail for `alifestd_mark_num_descendants_asexual`."""

    num_descendants = np.zeros_like(ancestor_ids)
    for idx_r, ancestor_id in enumerate(ancestor_ids[::-1]):
        idx = len(ancestor_ids) - 1 - idx_r
        if ancestor_id == idx:
            continue  # handle genesis cases

        num_descendants[ancestor_id] += num_descendants[idx] + 1

    return num_descendants


def alifestd_mark_num_descendants_asexual_slow_path(
    phylogeny_df: pd.DataFrame,
) -> pd.DataFrame:
    """Implementation detail for `alifestd_mark_num_descendants_asexual`."""

    phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["num_descendants"] = 0

    for idx in reversed(phylogeny_df.index):
        ancestor_id = phylogeny_df.at[idx, "ancestor_id"]
        if ancestor_id == idx:
            continue  # handle genesis cases

        own_num_descendants = phylogeny_df.at[idx, "num_descendants"]

        delta = own_num_descendants + 1
        phylogeny_df.at[ancestor_id, "num_descendants"] += delta

    return phylogeny_df


def alifestd_mark_num_descendants_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `num_descendants`, excluding self.

    A topological sort will be applied if `phylogeny_df` is not topologically
    sorted. Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df[
            "num_descendants"
        ] = alifestd_mark_num_descendants_asexual_fast_path(
            phylogeny_df["ancestor_id"].to_numpy()
        )
        return phylogeny_df
    else:
        return alifestd_mark_num_descendants_asexual_slow_path(phylogeny_df)



import contextlib
import logging
import math
import multiprocessing
import os
import typing
import uuid

from downstream import dataframe as dstream_dataframe
import pandas as pd
import polars as pl
import pyarrow as pa
import tqdm

from .._auxiliary_lib import (
    alifestd_make_empty,
    get_sole_scalar_value_polars,
    give_len,
    log_context_duration,
    log_memory_usage,
    render_polars_snapshot,
)
from ..phylogenetic_inference.tree._impl._build_tree_searchtable_cpp_impl_stub import (
    Records,
    collapse_unifurcations,
    extend_tree_searchtable_cpp_from_exploded,
    extract_records_to_dict,
)


def _sort_Tbar_argv(
    long_df: pl.DataFrame,
    num_slices: int,
    slice_index: int,
) -> pl.DataFrame:
    """Fast chronological sort within dstream data id groups, when Tbar_argv
    is available."""
    with log_context_duration(
        f"gather_indices ({slice_index + 1}/{num_slices})",
        logging.info,
    ):
        gather_indices = (  # argsort: what index should this row be sorted to?
            long_df.select(
                gather_indices=pl.when(  # where do data id's transition?
                    pl.col("dstream_data_id").shift(
                        fill_value=long_df["dstream_data_id"].first() + 1,
                    )
                    != pl.col("dstream_data_id")
                )
                .then(pl.int_range(len(long_df)))  # mark transition indices
                .otherwise(None)
                .forward_fill()  # fill first index across entire data id group
                .add(pl.col("dstream_Tbar_argv")),  # add Tbar_argv offset
            )
            .to_series()
            .to_numpy()
        )

    with log_context_duration(
        f".gather(gather_indices) ({slice_index + 1}/{num_slices})",
        logging.info,
    ):
        long_df = long_df.select(  # apply argsort
            pl.col(
                "dstream_data_id",
                "dstream_T",
                "dstream_Tbar",
                "dstream_value",
            ).gather(gather_indices),
        )

    return long_df


def _sort_Tbar(
    long_df: pl.DataFrame,
    num_slices: int,
    slice_index: int,
) -> pl.DataFrame:
    """Fallback chronological sort within dstream data id groups, when
    Tbar_argv is not available."""
    with log_context_duration(
        '.sort_by("dstream_Tbar").over(partition_by="dstream_data_id") '
        f"({slice_index + 1}/{num_slices})",
        logging.info,
    ):
        long_df = long_df.select(
            pl.col(
                "dstream_data_id",
                "dstream_T",
                "dstream_Tbar",
                "dstream_value",
            )
            .sort_by("dstream_Tbar")
            .over(partition_by="dstream_data_id"),
        )

    return long_df


def _make_exploded_slice(
    df_slice: pl.DataFrame,
    num_slices: int,
    slice_index: int,
) -> pl.DataFrame:
    """Explode dstream buffers to 1 differentia per row, calculating Tbar for
    each and ensuring strata are chronological within data id groups."""

    # explode dstream buffer to 1 differentia per row, calculating Tbar for each
    with log_context_duration(
        "dstream.dataframe.explode_lookup_unpacked "
        f"({slice_index + 1}/{num_slices})",
        logging.info,
    ):
        long_df = dstream_dataframe.explode_lookup_unpacked(
            df_slice, calc_Tbar_argv=True, value_type="uint64"
        )

    # ensure strata are chronological within data id groups
    # (i.e., within groups of differentia from same buffer)
    long_df = [  # dispatch to sort method based on available columns
        _sort_Tbar,
        _sort_Tbar_argv,
    ]["dstream_Tbar_argv" in long_df.columns](
        long_df=long_df,
        num_slices=num_slices,
        slice_index=slice_index,
    )

    if slice_index == 0:
        render_polars_snapshot(long_df, "exploded", logging.info)

    return long_df


def _produce_exploded_slices(
    queue: multiprocessing.JoinableQueue,
    df: typing.Union[pl.DataFrame, pl.LazyFrame],
    exploded_slice_size: int,
) -> None:
    """Exploded DataFrame in chunks, passed to queue for consumption."""
    with log_context_duration(
        "dstream.dataframe.unpack_data_packed", logging.info
    ):
        df = dstream_dataframe.unpack_data_packed(df)

    render_polars_snapshot(df, "unpacked", logging.info)

    # ensure genomes sorted by generations elapsed in ascending order
    # AFTER dstream_T has been unpacked, but before exploded
    with log_context_duration('.sort("dstream_T")', logging.info):
        df = df.sort("dstream_T", descending=False, maintain_order=True)

    render_polars_snapshot(df, "sorted", logging.info)

    if "dstream_data_id" not in df.columns:
        logging.info(" - adding dstream_data_id column")
        # ensure chunking doesn't affect data ids
        df = df.with_row_index("dstream_data_id")

    num_slices = math.ceil(len(df) / exploded_slice_size)
    logging.info(f"{len(df)=} {exploded_slice_size=} {num_slices=}")

    for slice_idx, df_slice in enumerate(df.iter_slices(exploded_slice_size)):
        logging.info(
            f"- worker exploding slice {slice_idx + 1} / {num_slices}"
        )
        # apply explode transformation
        long_df = _make_exploded_slice(
            df_slice=df_slice,
            num_slices=num_slices,
            slice_index=slice_idx,
        )

        # pass exploded data to consumer through queue via tmpfile
        # (better performing for fast read than passing directly through queue)
        logging.info("- worker putting exploded data")
        outpath = f"/tmp/{uuid.uuid4()}.arrow"
        long_df.select(pl.all().shrink_dtype()).write_ipc(
            outpath, compression="uncompressed"
        )
        del long_df  # clear memory
        queue.put(outpath)
        logging.info("- worker waiting for consumption")
        queue.join()  # wait for produced item to be consumed
        logging.info("- worker wait complete")

    logging.info("- worker putting sentinel value to signal completion")
    queue.put(None)  # send sentinel value to signal completion
    logging.info(" - worker complete")


def _build_records_chunked(
    slices: typing.Iterator[str],
    collapse_unif_freq: int,
    dstream_S: int,
    exploded_slice_size: int,
) -> Records:
    """Build tree searchtable from DataFrame, exploding in chunks to reduce
    memory usage."""
    init_size = exploded_slice_size * dstream_S * 2
    logging.info(f"{init_size=}")
    records = Records(init_size)  # handle for C++ tree-building data

    logging.info("consuming from exploded df worker")
    for i, inpath in enumerate(slices):
        logging.info(
            f"taking exploded df off queue ({i + 1} / {len(slices)})...",
        )

        logging.info(
            f"opening slice ({i + 1} / {len(slices)}) from {inpath}...",
        )
        with pa.memory_map(inpath, "rb") as source:  # use pyarrow fast reader
            pa_array = pa.ipc.open_file(source).read_all()

            logging.info(f"incorporating slice ({i + 1} / {len(slices)})...")
            with log_context_duration(
                "extend_tree_searchtable_cpp_from_exploded "
                f"({i + 1} / {len(slices)})",
                logging.info,
            ):
                # dispatch to C++ tree-building implementation
                extend_tree_searchtable_cpp_from_exploded(
                    records,
                    pa_array["dstream_data_id"].to_numpy(),
                    pa_array["dstream_T"].to_numpy(),
                    pa_array["dstream_Tbar"].to_numpy(),
                    pa_array["dstream_value"].to_numpy(),
                    tqdm.tqdm,
                )

        logging.info(f"unlinking slice {i + 1} / {len(slices)}...")
        os.unlink(inpath)

        if collapse_unif_freq > 0 and (i + 1) % collapse_unif_freq == 0:
            with log_context_duration(
                "collapse_unifurcations(dropped_only=True) "
                f"({i + 1} / {len(slices)})",
                logging.info,
            ):
                records = collapse_unifurcations(records, dropped_only=True)

        log_memory_usage(logging.info)

    logging.info("slices complete")

    # redundant w/ below (just here for testing)
    if collapse_unif_freq == -1:
        with log_context_duration(
            "collapse_unifurcations(dropped_only=True) (finalize)",
            logging.info,
        ):
            records = collapse_unifurcations(records, dropped_only=True)

    # collapse all unifs, to reduce subsequent memory pressure
    with log_context_duration(
        "collapse_unifurcations(dropped_only=False)",
        logging.info,
    ):
        records = collapse_unifurcations(records, dropped_only=False)

    log_memory_usage(logging.info)

    return records


def _join_user_defined_columns(
    df: pl.DataFrame, phylo_df: pl.DataFrame
) -> pl.DataFrame:
    """Join user-defined columns from input data onto reconstructed tree
    dataframe."""
    df = df.select(
        pl.exclude("^dstream_.*$", "^downstream_.*$"),
        pl.col("dstream_data_id").cast(pl.UInt64),
    )
    joined_columns = {*df.lazy().collect_schema().names()} - {
        *phylo_df.lazy().collect_schema().names()
    }
    if joined_columns:
        logging.info(f" - {len(joined_columns)} column(s) to join")
        logging.info(f" - joining columns: {[*joined_columns]}")
        phylo_df = phylo_df.join(
            df.lazy().collect(), on="dstream_data_id", how="left"
        )
    else:
        logging.info(" - no columns to join, skipping")

    return phylo_df


def _construct_result_dataframe(
    records: Records,
    differentia_bitwidth: int,
    dstream_S: int,
) -> pl.DataFrame:
    """Convert tree searchtable records to DataFrame."""
    logging.info("converting records to dict...")
    records_dict = extract_records_to_dict(records)

    logging.info("converting dict to dataframe...")
    schema = {
        "dstream_data_id": pl.UInt64,
        "id": pl.UInt64,
        "ancestor_id": pl.UInt64,
        "rank": pl.UInt64,
    }
    records_dict = {k: records_dict[k] for k in schema}
    return (
        pl.from_dict(
            records_dict,  # type: ignore
            schema=schema,
        )
        .with_columns(
            pl.lit(differentia_bitwidth)
            .alias("hstrat_differentia_bitwidth")
            .cast(pl.UInt32),
            pl.lit(dstream_S).alias("dstream_S").cast(pl.UInt32),
        )
        .rename({"rank": "hstrat_rank"})
    )


def _surface_unpacked_reconstruct(
    slices: typing.Iterator[str],
    *,
    collapse_unif_freq: int,
    differentia_bitwidth: int,
    dstream_S: int,
    exploded_slice_size: int,
) -> pl.DataFrame:
    """Reconstruct phylogenetic tree from unpacked dstream data."""
    logging.info("building tree searchtable chunkwise...")
    records = _build_records_chunked(
        slices,
        collapse_unif_freq=collapse_unif_freq,
        dstream_S=dstream_S,
        exploded_slice_size=exploded_slice_size,
    )

    with log_context_duration("_construct_result_dataframe", logging.info):
        phylo_df = _construct_result_dataframe(
            records,
            differentia_bitwidth=differentia_bitwidth,
            dstream_S=dstream_S,
        )

    del records  # clear memory
    render_polars_snapshot(phylo_df, "converted", logging.info)

    logging.info("surface_unpack_reconstruct complete")
    render_polars_snapshot(phylo_df, "reconstruction", logging.info)

    return phylo_df


@contextlib.contextmanager
def _generate_exploded_slices_mp(
    df: typing.Union[pl.LazyFrame, pl.DataFrame],
    exploded_slice_size: int,
    mp_context: str,
) -> typing.Iterator[typing.Iterator[str]]:
    """Generator wrapping genreation of exploded data frame slices via
    parallel multiprocess producer."""
    try:  # RE https://docs.pola.rs/user-guide/misc/multiprocessing/
        logging.info(f"attempting to use multiprocessing {mp_context} context")
        mp_context = multiprocessing.get_context(mp_context)
    except ValueError:  # forkserver available on unix only
        logging.info("attempting to use multiprocessing spawn context")
        mp_context = multiprocessing.get_context("spawn")

    logging.info("creating work queue")
    queue = mp_context.JoinableQueue()

    logging.info("spawning exploded df worker")
    producer = mp_context.Process(
        target=_produce_exploded_slices,
        args=(queue, df, exploded_slice_size),  # lazyframe is cheap to send
    )
    logging.info("starting exploded df worker")
    producer.start()

    num_slices = (
        df.lazy().select(pl.len()).collect().item() + (exploded_slice_size - 1)
    ) // exploded_slice_size

    yield give_len(  # enable len() on generator for nice logging
        # yield generated slices until sentinel value None is received,
        # immediately marking items as consumed (`task_done`) to trigger
        # the next item to be produced if queue has been emptied
        iter(lambda: (queue.get(), queue.task_done())[0], None),
        num_slices,
    )

    producer.join()  # wait for producer to finish (no effect, but good form)


def surface_unpack_reconstruct(
    df: typing.Union[pl.DataFrame, pl.LazyFrame],
    *,
    collapse_unif_freq: int = 1,
    exploded_slice_size: int = 1_000_000,
    mp_context: str = "spawn",
) -> pl.DataFrame:
    """Unpack dstream buffer and counter from genome data and construct an
    estimated phylogenetic tree for the genomes.

    Parameters
    ----------
    df : pl.DataFrame
        The input DataFrame containing packed data with required columns, one
        row per genome.

        Required schema:
            - 'data_hex' : pl.String
                - Raw genome data, with serialized dstream buffer and counter.
                - Represented as a hexadecimal string.
            - 'dstream_algo' : pl.Categorical
                - Name of downstream curation algorithm used.
                - e.g., 'dstream.steady_algo'
            - 'dstream_storage_bitoffset' : pl.UInt64
                - Position of dstream buffer field in 'data_hex'.
            - 'dstream_storage_bitwidth' : pl.UInt64
                - Size of dstream buffer field in 'data_hex'.
            - 'dstream_T_bitoffset' : pl.UInt64
                - Position of dstream counter field in 'data_hex'.
            - 'dstream_T_bitwidth' : pl.UInt64
                - Size of dstream counter field in 'data_hex'.
            - 'dstream_S' : pl.Uint32
                - Capacity of dstream buffer, in number of data items.

        Optional schema:
            - 'downstream_version' : pl.Categorical
                - Version of downstream library used to curate data items.
            - 'dstream_data_id' : pl.UInt64
                - Unique identifier for each data item.
                - If not provided, row number will be used as identifier.
            - 'downstream_exclude_exploded' : pl.Boolean
                - Should row be dropped after exploding unpacked data?
            - 'downstream_exclude_unpacked' : pl.Boolean
                - Should row be dropped after unpacking packed data?
            - 'downstream_validate_exploded' : pl.String, polars expression
                - Polars expression to validate exploded data.
            - 'downstream_validate_unpacked' : pl.String, polars expression
                - Polars expression to validate unpacked data.

    collapse_unif_freq : int, default 1
        Frequency of unifurcation collapse, in number of slices.

        Set to 0 to disable.

    exploded_slice_size : int, default 1_000_000
        Number of rows to process at once. Lower values reduce memory usage.

    mp_context : str, default 'spawn'
        Multiprocessing context to use for parallel processing.

    Returns
    -------
    pl.DataFrame
        The output DataFrame containing the estimated phylogenetic tree in
        alife standard format, with the following columns:

        - 'id' : pl.UInt64
            - Unique identifier for each taxon (RE alife standard format).
        - 'ancestor_id' : pl.UInt64
            - Unique identifier for ancestor taxon  (RE alife standard format).
        - 'hstrat_rank' : pl.UInt64
            - Num generations elapsed for ancestral differentia.
            - Corresponds to`dstream_Tbar` for inner nodes.
            - Corresponds `dstream_T` - 1 for leaf nodes
        - 'hstrat_differentia_bitwidth' : pl.UInt64
            - Size of annotation differentiae, in bits.
            - Corresponds to `dstream_value_bitwidth`.
        - 'dstream_S' : pl.UInt32
            - Capacity of dstream buffer, in number of data items.
        - 'dstream_data_id' : pl.UInt64
            - Unique identifier for each genome in source dataframe
            - Set to source dataframe row index if not provided.

        User-defined columns, except some prefixed with 'downstream_' or
        'dstream_', will be forwarded from the input DataFrame.

        Note that the alife-standard `ancestor_list` column is not included in
        the output.

    See Also
    --------
    surface_postprocess_trie :
        Post-processes raw phylogeny data, including collapsing superfluous
        internal nodes and estimating taxon origin times.
    alifestd_try_add_ancestor_list_col :
        Adds alife-standard `ancestor_list` column to phylogeny data.
    """
    logging.info("beginning surface_unpack_reconstruct")
    log_memory_usage(logging.info)

    render_polars_snapshot(df, "packed", logging.info)
    logging.info(f"packed {type(df)=}")

    # for simplicity, return early for this special case
    if df.lazy().limit(1).collect().is_empty():
        logging.info("empty input dataframe, returning empty result")
        res = alifestd_make_empty()
        res["taxon_label"] = None
        res["hstrat_rank"] = pd.Series(dtype=int)
        res["hstrat_differentia_bitwidth"] = pd.Series(dtype=int)
        res["dstream_S"] = pd.Series(dtype=int)
        return pl.from_pandas(res)

    logging.info("extracting metadata...")
    dstream_storage_bitwidth = get_sole_scalar_value_polars(
        df, "dstream_storage_bitwidth"
    )
    logging.info(f" - dstream storage bitwidth: {dstream_storage_bitwidth}")

    dstream_S = get_sole_scalar_value_polars(df, "dstream_S")
    logging.info(f" - dstream S: {dstream_S}")

    if not dstream_storage_bitwidth % dstream_S == 0:
        raise ValueError(
            "dstream_storage_bitwidth must be a multiple of dstream_S, "
            "cannot calculate differentia_bitwidth",
        )
    differentia_bitwidth = dstream_storage_bitwidth // dstream_S
    logging.info(f" - differentia bitwidth: {differentia_bitwidth}")

    logging.info("dispatching to surface_unpacked_reconstruct")
    with _generate_exploded_slices_mp(
        df, exploded_slice_size, mp_context
    ) as slices:
        phylo_df = _surface_unpacked_reconstruct(
            slices,
            collapse_unif_freq=collapse_unif_freq,
            differentia_bitwidth=differentia_bitwidth,
            dstream_S=dstream_S,
            exploded_slice_size=exploded_slice_size,
        )

    logging.info("joining user-defined columns...")
    with log_context_duration("_join_user_defined_columns", logging.info):
        try:
            phylo_df = _join_user_defined_columns(df, phylo_df)
        except pl.exceptions.ColumnNotFoundError:
            phylo_df = _join_user_defined_columns(
                df.with_row_index("dstream_data_id"),
                phylo_df,
            )

    return phylo_df



import types

from downstream import dstream, dsurf
import numpy as np
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "algo",
    [
        dstream.steady_algo,
        dstream.stretched_algo,
        dstream.tilted_algo,
        dstream.stretchedxtc_algo,
        dstream.tiltedxtc_algo,
        dstream.primed_0pad0_steady_algo,
        dstream.primed_0pad0_stretchedxtc_algo,
        dstream.primed_0pad0_tiltedxtc_algo,
    ],
)
@pytest.mark.parametrize("S", [8, 16, 32, np.zeros(32, dtype=np.uint64)])
@pytest.mark.parametrize("n", [0, 1, 5, 25, 100])
@pytest.mark.parametrize("dstream_T_bitwidth", [32, 64])
@pytest.mark.parametrize(
    "stratum_differentia_bit_width", [1, 2, 4, 8, 16, 32, 64]
)
def test_surf_to_hex_then_from_hex(
    algo: types.ModuleType,
    S: int,
    n: int,
    dstream_T_bitwidth: int,
    stratum_differentia_bit_width: int,
):
    surf = hstrat.HereditaryStratigraphicSurface(
        dsurf.Surface(algo, S),
        predeposit_strata=0,
        stratum_differentia_bit_width=stratum_differentia_bit_width,
    )
    surf.DepositStrata(n)
    assert (
        hstrat.surf_from_hex(
            hstrat.surf_to_hex(surf, dstream_T_bitwidth=dstream_T_bitwidth),
            dstream_algo=algo,
            dstream_T_bitwidth=dstream_T_bitwidth,
            dstream_storage_bitwidth=surf.S
            * surf.GetStratumDifferentiaBitWidth(),
            dstream_S=surf.S,
        )
        == surf
    )



import typing

import downstream
from downstream import dstream, dsurf
import polars as pl

from .._auxiliary_lib import alifestd_mark_leaves
from ..genome_instrumentation import HereditaryStratigraphicSurface
from ..serialization import surf_to_hex
from ..test_drive import descend_template_phylogeny_alifestd


def surface_test_drive(
    df: pl.LazyFrame,
    *,
    dstream_algo: str,
    dstream_S: int,
    dstream_T_bitwidth: int = 32,
    progress_wrap: typing.Callable = lambda x: x,
    stratum_differentia_bit_width: int,
) -> pl.DataFrame:
    """Reads alife standard phylogeny dataframe to create a population of
    hstrat surface annotations corresponding to the phylogeny tips, "as-if"
    they had evolved according to the provided phylogeny history.

    Parameters
    ----------
    df : pl.DataFrame
        The input DataFrame containing alife standard phylogeny with required
        columns, one row per taxon.

        Note that the alife-standard `ancestor_list` column is not required.

        Required schema:
            - 'id' : pl.UInt64
                - Taxon identifier.
            - 'ancestor_id' : pl.UInt64
                - Taxon identifier of ancestor.
                - Own 'id' if root.

        Optional schema:
            - 'origin_time' : pl.UInt64
                - Number of generations elapsed from ancestor.
                - Determines branch lengths.
                - Otherwise, all branches are assumed to be length 1.
            - 'extant' : pl.Boolean
                - Should an entry corresponding to this phylogeny taxon be
                  included in the output population?
                - Otherwise, all tips are considered extant and all inner nodes
                  are not.
            - Additional user-defined columns will be forwarded to the output
              DataFrame.

    dstream_algo : str
        Name of downstream curation algorithm to use.

    dstream_S : int
        Capacity of annotation dstream buffer, in number of data items.

    progress_wrap : Callable, optional
        Pass tqdm or equivalent to display a progress bar.

    stratum_differentia_bit_width : int
        The bit width of the generated differentia.

    Returns
    -------
    pl.DataFrame
        The output DataFrame containing generated hstrat surface annotations.

        Required schema:
            - 'data_hex' : pl.String
                - Raw genome data, with serialized dstream buffer and counter.
                - Represented as a hexadecimal string.
            - 'downstream_version' : pl.Categorical
                - Version of downstream library used.
            - 'dstream_algo' : pl.Categorical
                - Name of downstream curation algorithm used.
                - e.g., 'dstream.steady_algo'
            - 'dstream_storage_bitoffset' : pl.UInt32
                - Position of dstream buffer field in 'data_hex'.
            - 'dstream_storage_bitwidth' : pl.UInt32
                - Size of dstream buffer field in 'data_hex'.
            - 'dstream_T_bitoffset' : pl.UInt32
                - Position of dstream counter field in 'data_hex'.
            - 'dstream_T_bitwidth' : pl.UInt32
                - Size of dstream counter field in 'data_hex'.
            - 'dstream_S' : pl.Uint32
                - Capacity of dstream buffer, in number of data items.
            - 'origin_time' : pl.UInt64
                - Number of generations elapsed since the founding ancestor.
            - 'td_source_id' : pl.UInt64
                - Corresponding taxon identifier in source phylogeny.

        Additional user-defined columns will be forwarded from the input
        DataFrame.

    Notes
    -----
    - Input columns "id", "ancestor_id", and "ancestor_list" are not forwarded
      to output, to avoid conflicts with the output schema for subsequent
      phylogeny reconstruction.
    """
    df = df.lazy().collect()

    ancestor_instrument = HereditaryStratigraphicSurface(
        dstream_surface=dsurf.Surface(
            algo=eval(dstream_algo, {"dstream": dstream}),
            storage=dstream_S,
        ),
        predeposit_strata=0,
        stratum_differentia_bit_width=stratum_differentia_bit_width,
    )
    df_pd = df.to_pandas()
    if "extant" not in df_pd.columns:
        df_pd = alifestd_mark_leaves(df_pd, mutate=True)
        df_pd["extant"] = df_pd["is_leaf"].values

    extant_ids = df_pd.loc[df_pd["extant"].values, "id"].values
    surfaces = descend_template_phylogeny_alifestd(
        df_pd,
        extant_ids=extant_ids,
        seed_instrument=ancestor_instrument,
        progress_wrap=progress_wrap,
    )

    data = {
        "data_hex": [surf_to_hex(surf) for surf in surfaces],
        "downstream_version": downstream.__version__,
        "dstream_algo": dstream_algo,
        "dstream_storage_bitwidth": dstream_S * stratum_differentia_bit_width,
        "dstream_storage_bitoffset": dstream_T_bitwidth,
        "dstream_T_bitwidth": dstream_T_bitwidth,
        "dstream_T_bitoffset": 0,
        "dstream_S": dstream_S,
        "origin_time": [surf.GetNextRank() for surf in surfaces],
        "td_source_id": extant_ids,
    }
    schema = {
        "data_hex": pl.Utf8,
        "dstream_algo": pl.Categorical,
        "dstream_storage_bitwidth": pl.UInt32,
        "downstream_version": pl.Categorical,
        "dstream_storage_bitoffset": pl.UInt32,
        "dstream_T_bitwidth": pl.UInt32,
        "dstream_T_bitoffset": pl.UInt32,
        "dstream_S": pl.UInt32,
        "origin_time": pl.UInt64,
        "td_source_id": pl.UInt64,
    }
    return pl.concat(
        (
            pl.DataFrame(data, schema=schema),
            df.select(
                pl.exclude("id", "ancestor_id", "ancestor_list", *data.keys()),
            ).filter(df_pd["extant"].values),
        ),
        how="horizontal",
    )



import typing

if typing.TYPE_CHECKING:
    from ..frozen_instrumentation import HereditaryStratigraphicSpecimen
    from ..genome_instrumentation import (
        HereditaryStratigraphicColumn,
        HereditaryStratigraphicSurface,
    )

# using TypeVar as opposed to Union constrains multiple
# instances of the annotation in the same signature to be
# of the same type
# (e.g., must all be HereditaryStratigraphicColumn or
# all be HereditaryStratigraphicSurface)
HereditaryStratigraphicArtifact = typing.TypeVar(
    "HereditaryStratigraphicArtifact",  # name
    "HereditaryStratigraphicColumn",  # constraints
    "HereditaryStratigraphicSurface",
    "HereditaryStratigraphicSpecimen",
)



import os
import typing

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_collapse_unifurcations,
    alifestd_delete_unifurcating_roots_asexual,
    alifestd_make_empty,
    alifestd_splay_polytomies,
    alifestd_to_working_format,
    alifestd_unfurl_traversal_inorder_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    phylogeny_df = alifestd_splay_polytomies(phylogeny_df)
    phylogeny_df = alifestd_collapse_unifurcations(phylogeny_df)
    phylogeny_df = alifestd_delete_unifurcating_roots_asexual(phylogeny_df)
    original = phylogeny_df.copy()

    result = alifestd_unfurl_traversal_inorder_asexual(phylogeny_df)

    # Confirm that the input dataframe is not mutated.
    assert original.equals(phylogeny_df)

    # Verify that result is a permutation of the phylogeny IDs.
    assert len(result) == len(phylogeny_df)
    assert set(result) == set(phylogeny_df["id"])


def test_empty():
    res = alifestd_unfurl_traversal_inorder_asexual(alifestd_make_empty())
    assert len(res) == 0


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(apply: typing.Callable, mutate: bool):
    # Chain tree: 0 -> 1 -> 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_unfurl_traversal_inorder_asexual(
            phylogeny_df,
            mutate=mutate,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_inorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    valid_orders = {
        (3, 0, 4, 1, 2),
    }
    assert tuple(result) in valid_orders

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_inorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    valid_orders = {
        (3, 1, 4, 0, 5, 2, 6),
    }
    assert tuple(result.tolist()) in valid_orders

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple5(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_unfurl_traversal_inorder_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    valid_orders = {
        (3, 1, 5, 4, 6, 0, 2),
    }
    assert tuple(result.tolist()) in valid_orders

    if not mutate:
        assert original_df.equals(phylogeny_df)



import numpy as np
import pandas as pd

from ._alifestd_calc_clade_trait_frequency_asexual import (
    alifestd_calc_clade_trait_frequency_asexual,
)
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_mark_sister_asexual import alifestd_mark_sister_asexual
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_screen_trait_defined_clades_naive_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    mask_trait_absent: np.ndarray,  # 1D boolean array
    mask_trait_present: np.ndarray,  # 1D boolean array
    defining_mut_thresh: float = 0.75,
    defining_mut_sister_thresh: float = 0.75,
) -> np.ndarray:  # 1D boolean array
    """Perform a naive screen for trait-defined clades.

    This function checks if the trait frequency in a clade is above a certain
    threshold (`defining_mut_thresh`), and if the trait frequency in the sister
    clade is below a certain threshold (`defining_mut_sister_thresh`). Clades
    are defined as a node and all descendant nodes.

    The `mask_trait_absent` parameter can be used to exclude nodes from
    consideration, for instance &'ing with `alifestd_mark_leaves` can be used
    to only consider trait frequency among descendant leaves.

    Returns a numpy array of bool with the same length as the input DataFrame,
    with array elements as the number of nodes in the clade that have the
    trait. Returned array matches row order of the input DataFrame.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if "sister_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_sister_asexual(phylogeny_df, mutate=True)

    trait_frequency = alifestd_calc_clade_trait_frequency_asexual(
        phylogeny_df,
        mutate=True,
        mask_trait_absent=mask_trait_absent,
        mask_trait_present=mask_trait_present,
    )

    phylogeny_df[
        "alifestd_screen_trait_defined_clades_naive_asexual"
    ] = trait_frequency

    sister = phylogeny_df["sister_id"].values
    return (
        phylogeny_df[
            "alifestd_screen_trait_defined_clades_naive_asexual"
        ].values
        >= defining_mut_thresh
    ) & (
        phylogeny_df.loc[
            sister, "alifestd_screen_trait_defined_clades_naive_asexual"
        ].values
        <= defining_mut_sister_thresh
    )



import os

import pandas as pd
import pandas.testing as pdt
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    alifestd_collapse_unifurcations,
    alifestd_delete_trunk_asexual,
    alifestd_delete_trunk_asexual_polars,
    alifestd_prefix_roots,
    alifestd_test_leaves_isomorphic_asexual,
    alifestd_to_working_format,
    alifestd_validate,
)

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_no_trunk_column(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
        }
    )
    with pytest.raises(ValueError):
        alifestd_delete_trunk_asexual(df, mutate=mutate)
    with pytest.raises(ValueError):
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df))


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_single_trunk(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[]"],
            "is_trunk": [True],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_trunk_asexual(df, mutate=mutate)
    assert alifestd_validate(result)
    assert len(result) == 0

    if not mutate:
        assert original_df.equals(df)

    df.drop(columns=["ancestor_list"], inplace=True)
    result.drop(columns=["ancestor_list"], inplace=True)
    df["ancestor_id"] = [0]
    pdt.assert_frame_equal(
        result,
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df)).to_pandas(),
        check_dtype=False,
    )


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_no_collapse_needed1(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 8],
            "ancestor_id": [0, 0],
            "ancestor_list": ["[none]", "[0]"],
            "is_trunk": [True, False],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_trunk_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [8],
            "ancestor_list": ["[none]"],
            "is_trunk": [False],
        }
    )
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert original_df.equals(df)

    df.drop(columns=["ancestor_list"], inplace=True)
    result.drop(columns=["ancestor_list"], inplace=True)
    with pytest.raises(NotImplementedError):
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df)).to_pandas(),


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_no_collapse_needed2(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_list": ["[none]", "[0]"],
            "is_trunk": [True, True],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_trunk_asexual(df)
    assert alifestd_validate(result)
    assert len(result) == 0

    if not mutate:
        assert original_df.equals(df)

    df.drop(columns=["ancestor_list"], inplace=True)
    result.drop(columns=["ancestor_list"], inplace=True)
    df["ancestor_id"] = [0, 0]
    pdt.assert_frame_equal(
        result,
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df)).to_pandas(),
        check_dtype=False,
    )


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_no_collapse_needed3(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 1],
            "ancestor_list": ["[none]", "[0]", "[1]"],
            "is_trunk": [False, False, False],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_trunk_asexual(df)
    assert alifestd_validate(result)
    pd.testing.assert_frame_equal(result[df.columns], df)

    if not mutate:
        assert df.equals(original_df)

    df.drop(columns=["ancestor_list"], inplace=True)
    result.drop(columns=["ancestor_list"], inplace=True)
    pdt.assert_frame_equal(
        result,
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df)).to_pandas(),
        check_dtype=False,
    )


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_collapse1(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 4, 3, 5],
            "ancestor_id": [0, 0, 1, 2, 0, 5],
            "ancestor_list": ["[none]", "[0]", "[1]", "[2]", "[0]", "[none]"],
            "is_trunk": [True, True, True, False, False, False],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_trunk_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [4, 3, 5],
            "ancestor_id": [4, 3, 5],
            "ancestor_list": ["[none]", "[none]", "[none]"],
            "is_trunk": [False, False, False],
        }
    )
    pd.testing.assert_frame_equal(result[expected.columns], expected)

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_collapse2(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[1]",
                "[0]",
                "[0]",
                "[2]",
                "[3]",
            ],
            "is_trunk": [True, True, True, False, True, True, False],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_trunk_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [3, 6],
            "ancestor_list": ["[none]", "[3]"],
            "is_trunk": [False, False],
        }
    )
    pd.testing.assert_frame_equal(result[expected.columns], expected)

    if not mutate:
        assert df.equals(original_df)

    df.drop(columns=["ancestor_list"], inplace=True)
    result.drop(columns=["ancestor_list"], inplace=True)
    df["ancestor_id"] = [0, 0, 1, 0, 0, 2, 3]
    pdt.assert_frame_equal(
        result,
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df)).to_pandas(),
        check_dtype=False,
    )


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_trunk_asexual_noncontiguous_trunk(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "is_trunk": [True, False, True],
        }
    )
    original_df = df.copy()
    with pytest.raises(ValueError):
        alifestd_delete_trunk_asexual(df)

    if not mutate:
        assert df.equals(original_df)

    df.drop(columns=["ancestor_list"], inplace=True)
    df["ancestor_id"] = [0, 0, 1]
    with pytest.raises(ValueError):
        alifestd_delete_trunk_asexual_polars(pl.from_pandas(df))


def test_alifestd_delete_trunk_asexual_unifurcation():
    phylo = pd.read_csv(f"{assets}/trunktestphylo.csv")
    phylo["is_trunk"] = phylo["hstrat_rank"] < 64
    assert alifestd_test_leaves_isomorphic_asexual(
        phylo, phylo, taxon_label="dstream_data_id"
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        phylo,
        alifestd_collapse_unifurcations(phylo),
        taxon_label="dstream_data_id",
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        phylo,
        alifestd_collapse_unifurcations(phylo),
        taxon_label="dstream_data_id",
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        alifestd_delete_trunk_asexual(phylo),
        alifestd_delete_trunk_asexual(phylo),
        taxon_label="dstream_data_id",
    )
    assert alifestd_test_leaves_isomorphic_asexual(
        alifestd_collapse_unifurcations(alifestd_delete_trunk_asexual(phylo)),
        alifestd_delete_trunk_asexual(phylo),
        taxon_label="dstream_data_id",
    )

    pdt.assert_frame_equal(
        alifestd_delete_trunk_asexual(phylo),
        alifestd_delete_trunk_asexual_polars(
            pl.from_pandas(phylo)
        ).to_pandas(),
        check_dtype=False,
    )

    def clean(df: pd.DataFrame, allow_id_reassign: bool) -> pd.DataFrame:
        return alifestd_to_working_format(
            alifestd_collapse_unifurcations(
                alifestd_prefix_roots(
                    df,
                    allow_id_reassign=allow_id_reassign,
                    origin_time=64,
                ),
            ),
        )

    phylo["origin_time"] = phylo["hstrat_rank"]
    for allow_id_reassign in [True, False]:
        assert alifestd_test_leaves_isomorphic_asexual(
            clean(
                alifestd_delete_trunk_asexual(phylo),
                allow_id_reassign=allow_id_reassign,
            ),
            clean(
                alifestd_delete_trunk_asexual(
                    alifestd_collapse_unifurcations(phylo),
                ),
                allow_id_reassign=allow_id_reassign,
            ),
            taxon_label="dstream_data_id",
        )

        assert alifestd_test_leaves_isomorphic_asexual(
            clean(
                alifestd_delete_trunk_asexual(phylo),
                allow_id_reassign=allow_id_reassign,
            ),
            clean(
                alifestd_delete_trunk_asexual(
                    alifestd_collapse_unifurcations(phylo),
                ),
                allow_id_reassign=True,
            ),
            taxon_label="dstream_data_id",
        )

        assert alifestd_test_leaves_isomorphic_asexual(
            clean(
                alifestd_delete_trunk_asexual(phylo),
                allow_id_reassign=False,
            ),
            clean(
                alifestd_delete_trunk_asexual(
                    alifestd_collapse_unifurcations(phylo),
                ),
                allow_id_reassign=allow_id_reassign,
            ),
            taxon_label="dstream_data_id",
        )



import os
import typing

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_calc_clade_trait_count_asexual,
    alifestd_make_empty,
    alifestd_mark_leaves,
    alifestd_mark_num_leaves_asexual,
    alifestd_to_working_format,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    phylogeny_df = alifestd_make_empty()
    phylogeny_df["origin_time"] = None
    res = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        trait_mask=[],
    )
    assert len(res) == 0


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(apply: typing.Callable, mutate: bool):
    # Chain tree: 0 -> 1 -> 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=mutate,
        trait_mask=[False, True, True],
    )
    assert result[0] == 2
    assert result[1] == 2
    assert result[2] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
            "origin_time": [50, 20, 10, 30, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=mutate,
        trait_mask=[False, True, True, False, True],
    )
    assert result[0] == 0
    assert result[1] == 2
    assert result[2] == 3
    assert result[3] == 0
    assert result[4] == 1

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
            "origin_time": [0, 10, 20, 40, 30, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=mutate,
        trait_mask=[False, True, True, False, True, False, False],
    )
    assert result[0] == 3
    assert result[1] == 2
    assert result[2] == 1
    assert result[3] == 0
    assert result[4] == 1
    assert result[5] == 0
    assert result[6] == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple5(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0             7
    #       /   \          / \
    #      1     2        8   9
    #     / \
    #    3   4
    #       / \
    #      5   6
    #
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
                "[None]",
                "[7]",
                "[7]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=mutate,
        trait_mask=[
            False,
            True,
            True,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
        ],
    )
    assert result[0] == 3
    assert result[1] == 2
    assert result[2] == 1
    assert result[3] == 0
    assert result[4] == 1
    assert result[5] == 0
    assert result[6] == 0
    assert result[7] == 1
    assert result[8] == 0
    assert result[9] == 0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_fuzz(phylogeny_df: pd.DataFrame):
    phylogeny_df = alifestd_mark_leaves(phylogeny_df)
    phylogeny_df = alifestd_mark_num_leaves_asexual(phylogeny_df)
    original = phylogeny_df.copy()

    result = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df, trait_mask=phylogeny_df["is_leaf"]
    )
    assert (result == phylogeny_df["num_leaves"]).all()

    # Confirm that the input dataframe is not mutated.
    assert original.equals(phylogeny_df)



import typing

from astropy.utils.decorators import deprecated_renamed_argument
import dendropy as dp
import opytional as opyt

from ..._auxiliary_lib import (
    HereditaryStratigraphicInstrument,
    cast_int_lossless,
)
from ._descend_template_phylogeny import descend_template_phylogeny


@deprecated_renamed_argument("seed_column", "seed_instrument", since="1.20.0")
def descend_template_phylogeny_dendropy(
    tree: dp.Tree,
    seed_instrument: HereditaryStratigraphicInstrument,
    extant_nodes: typing.Optional[typing.Iterable[dp.Node]] = None,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicInstrument]:
    """Generate a population of hereditary stratigraphic instruments that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    tree : dendropy.Tree
        Phylogeny record as a dendropy Tree.
    seed_instrument : HereditaryStratigraphicInstrument
        Hereditary stratigraphic instrument to seed at root node of phylogeny.

        Returned hereditary stratigraphic instrument population will be
        generated as if repeatedly calling `CloneDescendant()` on
        `seed_instrument`. As such, specifies configuration (i.e., differentia
        bit width and stratum retention policy) for returned instruments. May
        already have strata deposited, which will be incorporated into generated
        extant population.
    extant_nodes : optional list of dendropy.Node
        Which organisms should hereditary stratigraphic instruments be created
        for?

        Designates content and order of returned list of hereditary
        stratigraphic instrument.

        If None, hereditary stratigraphic instruments will be created for all
        phylogenetic leaves (organisms without offspring) in order of
        appearance in `tree.leaf_node_iter()`.
    progress_wrap : Callable, optional
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicInstrument
        Population of hereditary stratigraphic instruments for extant lineage
        members (i.e., phylogeny leaf nodes).

        Instruments ordered in order of appearance of corresponding extant
        organism id.
    """

    return descend_template_phylogeny(
        ascending_lineage_iterables=(
            extant_node.ancestor_iter(
                inclusive=True,
            )
            for extant_node in opyt.or_value(
                extant_nodes, tree.leaf_node_iter()
            )
        ),
        descending_tree_iterable=tree.levelorder_node_iter(),
        get_parent=lambda node: node.parent_node,
        get_stem_length=lambda node: cast_int_lossless(
            node.edge_length, action="warn", context="edge length"
        ),
        seed_instrument=seed_instrument,
        demark=id,
        progress_wrap=progress_wrap,
    )



import os
import pathlib
import subprocess
import typing

import pytest

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_alifestd_as_newick_asexual_cli_help():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_as_newick_asexual",
            "--help",
        ],
        check=True,
    )


def test_alifestd_as_newick_asexual_cli_version():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_as_newick_asexual",
            "--version",
        ],
        check=True,
    )


@pytest.mark.parametrize(
    "input_file",
    [
        "nk_ecoeaselection.csv",
        "nk_lexicaseselection.csv",
        "nk_tournamentselection.csv",
    ],
)
@pytest.mark.parametrize(
    "taxon_label",
    [
        None,
        "id",
    ],
)
def test_alifestd_as_newick_asexual_cli_csv(
    input_file: str, taxon_label: typing.Optional[str]
):
    output_file = (
        "/tmp/hstrat-as_newick_asexual_cli-"
        f"{taxon_label}-{input_file}.newick"
    )
    pathlib.Path(output_file).unlink(missing_ok=True)
    cmd = [
        "python3",
        "-m",
        "hstrat._auxiliary_lib._alifestd_as_newick_asexual",
        "--input-file",
        f"{assets}/{input_file}",
        "-o",
        output_file,
    ]
    if taxon_label is not None:
        cmd.extend(["--taxon-label", taxon_label])
    subprocess.run(cmd, check=True)
    assert os.path.exists(output_file)



import math
import os
import typing

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_calc_clade_trait_frequency_asexual,
    alifestd_make_empty,
    alifestd_to_working_format,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    phylogeny_df = alifestd_make_empty()
    phylogeny_df["origin_time"] = None
    res = alifestd_calc_clade_trait_frequency_asexual(
        phylogeny_df,
        mask_trait_absent=[],
        mask_trait_present=[],
    )
    assert len(res) == 0


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(apply: typing.Callable, mutate: bool):
    # Chain tree: 0 -> 1 -> 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_frequency_asexual(
        phylogeny_df,
        mutate=mutate,
        mask_trait_absent=[True, False, False],
        mask_trait_present=[False, False, True],
    )
    assert result[0] == 0.5
    assert result[1] == 1.0
    assert result[2] == 1.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
            "origin_time": [50, 20, 10, 30, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_frequency_asexual(
        phylogeny_df,
        mutate=mutate,
        mask_trait_present=[False, False, False, False, True],
        mask_trait_absent=[True, False, False, True, False],
    )
    assert result[0] == 0.0
    assert result[1] == 0.5
    assert result[2] == pytest.approx(1 / 3)
    assert result[3] == 0.0
    assert result[4] == 1.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
            "origin_time": [0, 10, 20, 40, 30, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_frequency_asexual(
        phylogeny_df,
        mutate=mutate,
        mask_trait_absent=[False, False, False, False, True, False, False],
        mask_trait_present=[False, False, False, True, False, True, True],
    )
    assert result[0] == 0.75
    assert result[1] == 0.5
    assert result[2] == 1.0
    assert result[3] == 1.0
    assert result[4] == 0.0
    assert result[5] == 1.0
    assert result[6] == 1.0

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
def test_simple5(apply: typing.Callable, mutate: bool):
    # Tree structure:
    #         0             7
    #       /   \          / \
    #      1     2        8   9
    #     / \
    #    3   4
    #       / \
    #      5   6
    #
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
                "[None]",
                "[7]",
                "[7]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result = alifestd_calc_clade_trait_frequency_asexual(
        phylogeny_df,
        mutate=mutate,
        mask_trait_present=[
            False,
            False,
            True,
            True,
            False,
            True,
            True,
            True,
            False,
            False,
        ],
        mask_trait_absent=[
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
    )
    assert result[0] == pytest.approx(4 / 5)
    assert result[1] == 1.0
    assert result[2] == 1.0
    assert result[3] == 1.0
    assert result[4] == 1.0
    assert result[5] == 1.0
    assert result[6] == 1.0
    assert result[7] == 1.0
    assert math.isnan(result[8])
    assert math.isnan(result[9])

    if not mutate:
        assert original_df.equals(phylogeny_df)



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_mark_num_leaves_asexual import alifestd_mark_num_leaves_asexual
from ._alifestd_mark_sister_asexual import alifestd_mark_sister_asexual


def alifestd_mark_clade_leafcount_ratio_sister_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `clade_leafcount_ratio_sister`, containing the ratio of each
    clade's leaf count to that of its sister.

    Root nodes will have ratio 1. Tree must be strictly bifurcating.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    if "sister_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_sister_asexual(phylogeny_df, mutate=True)

    if "num_leaves" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_num_leaves_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["clade_leafcount_ratio_sister"] = (
        phylogeny_df["num_leaves"].values
    ) / (phylogeny_df.loc[phylogeny_df["sister_id"], "num_leaves"].values)

    return phylogeny_df



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_mark_clade_duration_asexual import (
    alifestd_mark_clade_duration_asexual,
)
from ._alifestd_mark_sister_asexual import alifestd_mark_sister_asexual


def alifestd_mark_clade_duration_ratio_sister_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `clade_duration_ratio_sister`, containing the ratio of each
    clade's duration to that of its sister.

    Root nodes will have ratio 1, unless also a leaf node. Leaf nodes and
    leaf-sisters may have ratio inf or NaN.

    Tree must be strictly bifurcating.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    if "sister_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_sister_asexual(phylogeny_df, mutate=True)

    if "clade_duration" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_clade_duration_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["clade_duration_ratio_sister"] = (
        phylogeny_df["clade_duration"].values
    ) / (phylogeny_df.loc[phylogeny_df["sister_id"], "clade_duration"].values)

    return phylogeny_df



import logging

import pandas as pd
import polars as pl

from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_delete_trunk_asexual_polars(
    phylogeny_df: pd.DataFrame,
) -> pd.DataFrame:
    """Delete entries masked by `is_trunk` column.

    Masked entries must be contiguous, meaning that no non-trunk entry can
    be an ancestor of a trunk entry. Children of deleted entries will become
    roots.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_collapse_trunk_asexual
    """
    if "is_trunk" not in phylogeny_df:
        raise ValueError(
            "`is_trunk` column not provided, trunk is unspecified"
        )

    if "ancestor_list" in phylogeny_df:
        raise NotImplementedError

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    has_contiguous_ids = phylogeny_df.select(
        pl.col("id").diff() == 1
    ).to_series().all() and (phylogeny_df["id"].first() == 0)
    if not has_contiguous_ids:
        raise NotImplementedError("non-contiguous ids not yet supported")

    logging.info(
        "- alifestd_delete_trunk_asexual_polars: marking ancestor_is_trunk...",
    )
    phylogeny_df = phylogeny_df.with_columns(
        ancestor_is_trunk=pl.col("is_trunk").gather(
            pl.col("ancestor_id"),
        )
    )

    logging.info(
        "- alifestd_delete_trunk_asexual_polars: testing special cases..."
    )
    if (phylogeny_df["is_trunk"] & ~phylogeny_df["ancestor_is_trunk"]).any():
        raise ValueError("specified trunk is non-contiguous")

    if phylogeny_df["is_trunk"].sum() == 0:
        return phylogeny_df

    logging.info(
        "- alifestd_delete_trunk_asexual_polars: updating ancestor_id...",
    )
    phylogeny_df = phylogeny_df.with_columns(
        ancestor_id=pl.when(pl.col("ancestor_is_trunk"))
        .then(pl.col("id"))
        .otherwise(pl.col("ancestor_id")),
    )

    logging.info("- alifestd_delete_trunk_asexual_polars: filtering...")
    phylogeny_df = phylogeny_df.filter(~pl.col("is_trunk"))

    assert phylogeny_df["is_trunk"].sum() == 0
    return phylogeny_df



import argparse
import logging
import os

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._alifestd_mark_oldest_root import alifestd_mark_oldest_root
from ._alifestd_mark_roots import alifestd_mark_roots
from ._configure_prod_logging import configure_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


def alifestd_join_roots(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """Point all other roots to oldest root, measured by lowest `origin_time`
    (if available) or otherwise lowest `id`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_roots(phylogeny_df, mutate=True)
    phylogeny_df = alifestd_mark_oldest_root(phylogeny_df, mutate=True)

    if len(phylogeny_df) <= 1:
        return phylogeny_df

    global_root_id = phylogeny_df.loc[
        phylogeny_df["is_oldest_root"].idxmax(), "id"
    ]

    if "ancestor_id" in phylogeny_df:
        phylogeny_df.loc[
            phylogeny_df["is_root"], "ancestor_id"
        ] = global_root_id

    if "ancestor_list" in phylogeny_df:
        phylogeny_df.loc[
            phylogeny_df["is_root"], "ancestor_list"
        ] = f"[{global_root_id}]"
        phylogeny_df.loc[
            phylogeny_df["is_oldest_root"], "ancestor_list"
        ] = "[none]"

    phylogeny_df["is_root"] = False
    phylogeny_df.loc[phylogeny_df["is_oldest_root"], "is_root"] = True

    return phylogeny_df


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Point all other roots to oldest root, measured by lowest `origin_time` (if available) or otherwise lowest `id`.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_join_roots",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    configure_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_join_roots", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_join_roots,
            ),
            overridden_arguments="ignore",  # seed is overridden
        )



import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_max_descendant_origin_time_asexual import (
    alifestd_mark_max_descendant_origin_time_asexual,
)


def alifestd_mark_clade_subtended_duration_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `clade_subtended_duration`, containing the difference between each the
    `origin_time` of each node's ancestor and the maximum
    `origin_time` of its descendants.

    Ancestor origin time for root nodes will be 0.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if "origin_time" not in phylogeny_df.columns:
        raise ValueError("phylogeny_df must contain `origin_time` column")

    if "max_descendant_origin_time" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_max_descendant_origin_time_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    ancestor_origin_times = np.where(
        phylogeny_df["ancestor_id"].values != phylogeny_df["id"].values,
        phylogeny_df.loc[
            phylogeny_df["ancestor_id"].values, "origin_time"
        ].values,
        0,
    )

    phylogeny_df["clade_subtended_duration"] = (
        phylogeny_df["max_descendant_origin_time"].values
        - ancestor_origin_times
    )

    return phylogeny_df



import pandas as pd
import pandas.testing as pdt
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_empty,
    alifestd_prefix_roots,
    alifestd_prefix_roots_polars,
)


def test_empty_df():
    df = alifestd_make_empty()
    result = alifestd_prefix_roots(df)
    assert result.empty

    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(df))


@pytest.mark.parametrize("mutate", [False, True])
def test_single_root_with_origin_time(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "origin_time": [10],
            "ancestor_id": [0],
            "ancestor_list": ["[none]"],
        }
    )
    original = df.copy()

    result = alifestd_prefix_roots(df, origin_time=5, mutate=mutate)
    if not mutate:
        pdt.assert_frame_equal(df, original)

    expected = pd.DataFrame(
        {
            "id": [0, 1],
            "origin_time": [10, 5],
            "ancestor_id": [1, 1],
            "ancestor_list": ["[1]", "[]"],
        }
    )
    pdt.assert_frame_equal(
        result.reset_index(drop=True), expected, check_like=True
    )

    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original))
    df.drop(columns=["ancestor_list"], inplace=True)
    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original), origin_time=5)


@pytest.mark.parametrize("mutate", [False, True])
def test_single_root_without_origin_time(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "origin_time": [10],
            "ancestor_id": [0],
            "ancestor_list": ["[none]"],
        }
    )
    original = df.copy()
    result = alifestd_prefix_roots(df, origin_time=None, mutate=mutate)
    if not mutate:
        pdt.assert_frame_equal(df, original)

    expected = pd.DataFrame(
        {
            "id": [0, 1],
            "origin_time": [10, 0],
            "ancestor_id": [1, 1],
            "ancestor_list": ["[1]", "[]"],
        }
    )
    pdt.assert_frame_equal(
        result.reset_index(drop=True), expected, check_like=True
    )

    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original))
    df.drop(columns=["ancestor_list"], inplace=True)
    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(
            pl.from_pandas(original), origin_time=None
        )


@pytest.mark.parametrize("mutate", [False, True])
def test_multiple_roots(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "origin_time": [20, 15, 10, 5],
            "ancestor_id": [0, 0, 2, 2],
            "ancestor_list": ["[none]", "[0]", "[none]", "[2]"],
        }
    )
    original = df.copy()
    result = alifestd_prefix_roots(df, origin_time=12, mutate=mutate)
    if not mutate:
        pdt.assert_frame_equal(df, original)

    expected = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "origin_time": [20, 15, 10, 5, 12],
            "ancestor_id": [4, 0, 2, 2, 4],
            "ancestor_list": ["[4]", "[0]", "[none]", "[2]", "[]"],
        }
    )
    pdt.assert_frame_equal(
        result.reset_index(drop=True), expected, check_like=True
    )

    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original))
    df.drop(columns=["ancestor_list"], inplace=True)
    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original), origin_time=12)


@pytest.mark.parametrize("mutate", [False, True])
def test_no_eligible_roots(mutate: bool):

    df = pd.DataFrame(
        {
            "id": [0, 1],
            "origin_time": [10, 5],
            "ancestor_id": [0, 0],
            "ancestor_list": ["[none]", "[0]"],
        }
    )
    original = df.copy()
    result = alifestd_prefix_roots(df, origin_time=15, mutate=mutate)
    if not mutate:
        pdt.assert_frame_equal(df, original)

    pdt.assert_frame_equal(
        result.reset_index(drop=True),
        original.reset_index(drop=True),
        check_like=True,
    )

    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original))
    df.drop(columns=["ancestor_list"], inplace=True)
    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original), origin_time=15)


@pytest.mark.parametrize("mutate", [False, True])
def test_warn_on_origin_time_delta(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "origin_time": [10],
            "origin_time_delta": [5],
            "ancestor_id": [0],
            "ancestor_list": ["[none]"],
        }
    )
    original = df.copy()
    with pytest.warns(UserWarning):
        alifestd_prefix_roots(df, origin_time=5, mutate=mutate)
    if not mutate:
        pdt.assert_frame_equal(df, original)

    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original))
    df.drop(columns=["ancestor_list"], inplace=True)
    with pytest.raises(NotImplementedError):
        alifestd_prefix_roots_polars(pl.from_pandas(original), origin_time=5)


@pytest.mark.parametrize("mutate", [False, True])
def test_fast_path_single_root(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "origin_time": [10, 9, 8],
            "ancestor_id": [0, 0, 1],
            "is_root": [True, False, False],
        },
    )
    original = df.copy()
    result = alifestd_prefix_roots(
        df,
        allow_id_reassign=True,
        origin_time=5,
        mutate=mutate,
    )
    if not mutate:
        pdt.assert_frame_equal(df, original)

    expected = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "origin_time": [5, 10, 9, 8],
            "ancestor_id": [0, 0, 1, 2],
        }
    )
    pdt.assert_frame_equal(
        result.reset_index(drop=True), expected, check_like=True
    )

    pdt.assert_frame_equal(
        result.reset_index(drop=True),
        alifestd_prefix_roots_polars(
            pl.from_pandas(original), allow_id_reassign=True, origin_time=5
        ).to_pandas(),
        check_dtype=False,
    )


@pytest.mark.parametrize("mutate", [False, True])
def test_fast_path_multiple_roots(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "origin_time": [15, 22, 29, 28, 8, 9, 10],
            "ancestor_id": [0, 0, 1, 1, 4, 4, 4],
        },
    )
    original = df.copy()
    result = alifestd_prefix_roots(
        df,
        allow_id_reassign=True,
        origin_time=10,
        mutate=mutate,
    )
    if not mutate:
        pdt.assert_frame_equal(df, original)

    expected = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7],
            "origin_time": [10, 15, 22, 29, 28, 8, 9, 10],
            "ancestor_id": [0, 0, 1, 2, 2, 5, 5, 5],
        },
    )
    pdt.assert_frame_equal(
        result.reset_index(drop=True), expected, check_like=True
    )

    pdt.assert_frame_equal(
        result.reset_index(drop=True),
        alifestd_prefix_roots_polars(
            pl.from_pandas(original), allow_id_reassign=True, origin_time=10
        ).to_pandas(),
        check_dtype=False,
    )


@pytest.mark.parametrize("mutate", [False, True])
def test_fast_path_no_eligible_roots(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "origin_time": [5, 4, 3],
            "ancestor_id": [0, 0, 1],
        }
    )
    original = df.copy()
    result = alifestd_prefix_roots(
        df,
        allow_id_reassign=True,
        origin_time=10,
        mutate=mutate,
    )
    if not mutate:
        pdt.assert_frame_equal(df, original)
    pdt.assert_frame_equal(result, original)

    pdt.assert_frame_equal(
        result.reset_index(drop=True),
        alifestd_prefix_roots_polars(
            pl.from_pandas(original), allow_id_reassign=True, origin_time=10
        ).to_pandas(),
        check_dtype=False,
    )



#!/usr/bin/env python3

import argparse
import functools
import gc
import os
import random
import sys
import types
import typing
import uuid

import downstream
from downstream import dstream
import numpy as np
import opytional as opyt
import pandas as pd
from tqdm import tqdm

from hstrat._auxiliary_lib import RngStateContext

try:
    from phylotrackpy import systematics
except (ImportError, ModuleNotFoundError) as e:
    print("phylotrackpy required for dstream surf evolution example")
    print("python3 -m pip install phylotrackpy")
    raise e

evolution_selector = random.Random(1)  # ensure consistent true phylogeny


def consistent_state_randrange(bitwidth: int):
    """A wrapper for randrange that preserves state regardless of bitwidth

    Motivation: the random state after randrange(2**64) was different
    than that of randrange(2**8), and we wanted to keep things like
    extracted fossils and generated UUIDs consistent between runs
    regardless of parameters like `differentia_bitwidth`. Therefore,
    we use this to avoid random state changing differently across runs.
    """
    return random.randrange(2**64) & ((1 << bitwidth) - 1)


def make_uuid4_fast(
    random_generator: typing.Union[types.ModuleType, random.Random]
) -> str:
    """Fast UUID4 generator, using lower-quality randomness."""
    return str(
        uuid.UUID(
            int=random_generator.getrandbits(128),
            version=4,
        )
    )


def extract_fossils(
    pop: typing.List,
    fossil_sample_proportion: float = 0.1,
) -> typing.List:
    """
    We extract CreateOffspring() instead of the parent itself, becauase
    parents with surviving children are not treated as leaf nodes by
    phylotrackpy; this simplifies true/reconst phylo comparison
    """
    return [
        parent.CreateOffspring(fossil=True)
        for parent in random.sample(
            pop,
            k=int(len(pop) * fossil_sample_proportion),
        )
    ]


def evolve_drift(
    population: typing.List,
    fossil_interval: typing.Optional[int] = None,
) -> typing.List:
    """
    Simple asexual evolutionary algorithm under drift conditions.
    Fossils refer to organisms saved in the middle of the process
    of evolution, when normally they would have been deleted for
    the next generation. Fossils supplement the final population
    to add a sampling of historical, extinct taxa as phylogeny tips.
    """
    # synchronous generations
    fossils = []
    for generation in tqdm(range(500)):
        population = [
            parent.CreateOffspring()
            for parent in evolution_selector.choices(
                population, k=len(population)
            )
        ]
        if fossil_interval and generation % fossil_interval == 0:
            with RngStateContext(random.randint(1, sys.maxsize)):
                fossils.extend(extract_fossils(population))

    # asyncrhonous generations
    nsplit = len(population) // 2
    for generation in tqdm(range(500)):
        population[:nsplit] = [
            parent.CreateOffspring()
            for parent in evolution_selector.choices(
                population[:nsplit], k=nsplit
            )
        ]
        if fossil_interval and generation % fossil_interval == 0:
            with RngStateContext(random.randint(1, sys.maxsize)):
                fossils.extend(extract_fossils(population))
        evolution_selector.shuffle(population)

    return [*fossils, *population]


def make_Organism(
    dstream_algo: types.ModuleType,
    differentia_bitwidth: int,
    surface_size: int,
    syst: systematics.Systematics,
) -> typing.Type:
    """
    Factory function to configure Organism class using `differentia_bitwidth`
    and `surface_size`, with `syst` to track phylogenies.
    """

    surf_dtype = {
        1: np.uint8,
        8: np.uint8,
        64: np.uint64,
    }[differentia_bitwidth]
    empty_surface = np.empty(surface_size, dtype=surf_dtype)
    empty_surface.setflags(write=False)

    surface_bitwidth = differentia_bitwidth * surface_size

    assign_site = functools.lru_cache(dstream_algo.assign_storage_site)
    algo_name = dstream_algo.__name__.split(".")[-1]

    class Organism:
        """Simple organism class, with instrumentation for both phylotrackpy
        and hstrat surface phylogeny tracking."""

        __slots__ = [
            "trait",
            "uid",
            "taxon",
            "dstream_T",
            "hstrat_surface",
            "is_fossil",
        ]

        # primary simulation business --- arbitrary data in this simple example
        trait: float

        # instrumentation: perfect tracking with phylotrackpy
        uid: str
        taxon: systematics.Taxon

        # instrumentation: approximate tracking with hstrat surface
        dstream_T: int
        hstrat_surface: np.ndarray

        # keep track of whether or not this Organism is a fossil
        is_fossil: bool

        @staticmethod
        def create_founder() -> "Organism":
            """Create a founder organism, with hstrat surface initailized."""
            founder = None
            with RngStateContext(random.randint(1, sys.maxsize)):
                for T in range(surface_size):
                    founder = opyt.apply_if_or_else(
                        founder,
                        lambda x: x.CreateOffspring(fossil=True),
                        Organism,
                    )
            assert founder.dstream_T == surface_size
            return founder

        def __init__(
            self: "Organism",
            parent_taxon: typing.Optional[systematics.Taxon] = None,
            parent_hstrat_surface: np.ndarray = empty_surface,
            parent_dstream_T: int = 0,
            trait: float = 0.0,
            is_fossil: bool = False,
        ) -> None:
            """Initialize organism, by default as root organism."""
            # handle primary simulation business...
            self.trait = (
                trait  # arbitrary example data, not used in simulation
            )

            # handle phylotrackpy instrumentation...
            self.uid = make_uuid4_fast(
                random if is_fossil else evolution_selector
            )
            self.taxon = syst.add_org(self, parent_taxon)

            # handle hstrat surface instrumentation...
            self.hstrat_surface = parent_hstrat_surface.copy()
            # ... deposit stratum...
            # hack to avoid varying ranges give different random states
            differentia_value = consistent_state_randrange(
                bitwidth=differentia_bitwidth
            )
            self.DepositStratum(differentia_value, parent_dstream_T)
            self.dstream_T = parent_dstream_T + 1
            self.is_fossil = is_fossil

        def __del__(self: "Organism") -> None:
            """Remove organism from phylotrackpy systematics."""
            syst.remove_org(self.taxon)

        def CreateOffspring(
            self: "Organism", *, fossil: bool = False
        ) -> "Organism":
            """Create an offspring organism, with mutation and
            generation-updated instrumentation."""
            return Organism(
                parent_taxon=self.taxon,
                parent_hstrat_surface=self.hstrat_surface.copy(),
                parent_dstream_T=self.dstream_T,
                trait=self.trait + np.random.uniform(-1, 1),
                is_fossil=fossil,
            )

        def DepositStratum(
            self: "Organism",
            differentia_value: int,
            dstream_T: int,
        ) -> None:
            assert dstream_algo.has_ingest_capacity(
                surface_size, dstream_T + 1
            )

            dstream_site = assign_site(surface_size, dstream_T)
            if dstream_site is not None:  # handle skip/discard case
                self.hstrat_surface[dstream_site] = differentia_value

        def ToHex(self: "Organism") -> str:
            """Serialize the organism to a hex string, for genome output."""
            T_arr = np.asarray(self.dstream_T, dtype=np.uint32)
            T_bytes = T_arr.astype(">u4").tobytes()  # big-endian u32
            T_hex = T_bytes.hex()

            differentia_bytewidth = differentia_bitwidth // 8
            pack_op = [
                lambda x: x.astype(f">u{differentia_bytewidth}"),  # big-endian
                np.packbits,  # default big bitorder
            ][differentia_bitwidth == 1]
            surface_bits = pack_op(self.hstrat_surface)
            surface_bytes = surface_bits.tobytes()
            surface_hex = surface_bytes.hex()

            trait_arr = np.asarray(self.trait, dtype=np.float32)
            trait_bytes = trait_arr.tobytes()
            trait_hex = trait_bytes.hex()

            return T_hex + surface_hex + trait_hex

        def ToRecord(self: "Organism") -> dict:
            """Serialize the organism to a dictionary."""
            return {
                "downstream_version": downstream.__version__,
                "data_hex": self.ToHex(),
                "taxon_label": self.uid,
                "dstream_algo": f"dstream.{algo_name}",
                "dstream_storage_bitoffset": 32,
                "dstream_storage_bitwidth": surface_bitwidth,
                "dstream_T_bitoffset": 0,
                "dstream_T_bitwidth": 32,
                "dstream_S": surface_size,
                "trait": self.trait,
                "is_fossil": self.is_fossil,
            }

    return Organism


def make_validation_record(
    Organism: typing.Type,
    n_gen: int,
    differentia_override: typing.Callable,
    validator_exploded: str,
    validator_unpacked: str,
) -> dict:
    """
    Generate ephemeral validation data for quality assurance. This
    helps ensure data is being written and read in the proper format
    in case of different endianness or other scenarios.
    """
    organism = None
    with RngStateContext(random.randint(1, sys.maxsize)):
        for T in range(n_gen):
            organism = opyt.apply_if_or_else(
                organism,
                Organism.CreateOffspring,
                Organism,
            )
            organism.DepositStratum(differentia_override(T), T)

    return {
        **organism.ToRecord(),
        "downstream_exclude_exploded": True,
        "downstream_validate_exploded": validator_exploded,
        "downstream_validate_unpacked": validator_unpacked,
    }


def _parse_args() -> argparse.Namespace:
    """Helper function to parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--differentia-bitwidth", type=int, default=1)
    parser.add_argument("--surface-size", type=int, default=64)
    parser.add_argument(
        "--genome-df-path",
        type=str,
        default="/tmp/genome-evolve_surf_dstream.pqt",
    )
    parser.add_argument(
        "--phylo-df-path",
        type=str,
        default="/tmp/phylo-evolve_surf_dstream.csv",
    )
    parser.add_argument(
        "--fossil-interval",
        type=int,
    )

    args = parser.parse_args()

    if args.differentia_bitwidth not in (1, 8, 64):
        raise NotImplementedError()

    if args.surface_size < 8:
        raise NotImplementedError()

    if not args.phylo_df_path.endswith(".csv"):
        raise ValueError("phylo_df_path must end with '.csv'")

    return args


def _get_df_save_handler(path: str) -> typing.Callable:
    """Helper function to get the appropriate pandas dataframe save handler."""
    df_save_handlers = {
        ".csv": pd.DataFrame.to_csv,
        ".json": pd.DataFrame.to_json,
        ".pqt": pd.DataFrame.to_parquet,
        ".parquet": pd.DataFrame.to_parquet,
    }
    ext = os.path.splitext(path)[-1]
    try:
        return df_save_handlers[ext]
    except KeyError:
        raise ValueError(f"Unsupported dataframe path extension: {ext}")


if __name__ == "__main__":
    np.random.seed(2)  # ensure reproducibility
    random.seed(2)

    args = _parse_args()

    # configure organism class
    syst = systematics.Systematics(lambda x: x.uid)  # each org is own taxon
    syst.add_snapshot_fun(systematics.Taxon.get_info, "taxon_label")
    Organism = make_Organism(
        dstream_algo=dstream.steady_algo,
        differentia_bitwidth=args.differentia_bitwidth,
        surface_size=args.surface_size,
        syst=syst,
    )

    # do simulation
    common_ancestor = Organism.create_founder()
    init_population = [common_ancestor.CreateOffspring() for _ in range(100)]
    sampled_genomes = evolve_drift(
        init_population, fossil_interval=args.fossil_interval
    )

    # mark non-tip taxa extinct
    del common_ancestor
    del init_population
    gc.collect()

    print("num organisms retained in exact tracker:", syst.get_total_orgs())
    print("final population size:", len(sampled_genomes))

    # set up validators to test during downstream processing
    S = args.surface_size
    checkerboard_validator = (
        "pl.col('dstream_value') == pl.col('dstream_Tbar') % 2"
    )
    block_validator = (
        f"pl.col('dstream_value') == pl.col('dstream_Tbar') // {S // 2 + 1}"
    )

    # write out the final population, including hstrat surface data
    genome_records = [
        *map(Organism.ToRecord, sampled_genomes),  # experiment data
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S,
            differentia_override=lambda T: T % 2,
            validator_exploded=checkerboard_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S}",
        ),
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S + 1,
            differentia_override=lambda T: T % 2,
            validator_exploded=checkerboard_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S + 1}",
        ),
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S,
            differentia_override=lambda T: T // (S // 2 + 1),
            validator_exploded=block_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S}",
        ),
        make_validation_record(  # ephemeral validation data
            Organism=Organism,
            n_gen=S + 1,
            differentia_override=lambda T: T // (S // 2 + 1),
            validator_exploded=block_validator,
            validator_unpacked=f"pl.col('dstream_T') == {S + 1}",
        ),
    ]
    genome_df = pd.DataFrame(genome_records)
    _get_df_save_handler(args.genome_df_path)(genome_df, args.genome_df_path)

    # write out ground truth phylogeny
    syst.snapshot(args.phylo_df_path)



from fishersrc import pvalue_npy as fisher_pvalue_npy
import numpy as np
import pandas as pd

from ._alifestd_calc_clade_trait_count_asexual import (
    alifestd_calc_clade_trait_count_asexual,
)
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_mark_sister_asexual import alifestd_mark_sister_asexual
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_screen_trait_defined_clades_fisher_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    mask_trait_absent: np.ndarray,  # 1D boolean array
    mask_trait_present: np.ndarray,  # 1D boolean array
) -> np.ndarray:  # 1D float array
    """Perform a screen for trait-defined clades based on Fisher's exact test.

    This function computes a Fisher's exact test comparing the trait frequency
    (number of clade members with the trait, number of clade members without
    the trait) in a clade with its sister clade. Returned values are one-tailed
    p-values for the hypothesis that the trait frequency in the clade is
    greater than in the sister clade.

    Root clades will be compared to themselves, as they have no sister clade.
    As such, root clades will take on p-values > 0.5.

    The `mask_trait_absent` parameter can be used to exclude nodes from
    consideration, for instance &'ing with `alifestd_mark_leaves` can be used
    to only consider trait frequency among descendant leaves.

    Returns a numpy array of bool with the same length as the input DataFrame,
    with array elements as the number of nodes in the clade that have the
    trait. Returned array matches row order of the input DataFrame.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if "sister_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_sister_asexual(phylogeny_df, mutate=True)

    trait_count_absent = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=True,
        trait_mask=mask_trait_absent,
    ).astype(np.uint32)

    trait_count_present = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=True,
        trait_mask=mask_trait_present,
    ).astype(np.uint32)

    phylogeny_df[
        "alifestd_screen_trait_defined_clades_fisher_asexual_absent"
    ] = trait_count_absent
    phylogeny_df[
        "alifestd_screen_trait_defined_clades_fisher_asexual_present"
    ] = trait_count_present

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.set_index("id", drop=False, inplace=True)

    sister = phylogeny_df["sister_id"].values
    sister_trait_count_absent = phylogeny_df.loc[
        sister, "alifestd_screen_trait_defined_clades_fisher_asexual_absent"
    ].values
    sister_trait_count_present = phylogeny_df.loc[
        sister, "alifestd_screen_trait_defined_clades_fisher_asexual_present"
    ].values

    sparse_mask = trait_count_present != 0

    sparse_result = fisher_pvalue_npy(
        a_true=trait_count_present[sparse_mask],
        a_false=trait_count_absent[sparse_mask],
        b_true=sister_trait_count_present[sparse_mask],
        b_false=sister_trait_count_absent[sparse_mask],
    )[
        1
    ]  # one-tailed p-value

    result = np.ones_like(mask_trait_absent, dtype=float)
    result[sparse_mask] = sparse_result

    return result



import math
import os
import typing

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_collapse_unifurcations,
    alifestd_delete_unifurcating_roots_asexual,
    alifestd_make_empty,
    alifestd_mark_clade_logistic_growth_children_asexual,
    alifestd_mark_leaves,
    alifestd_splay_polytomies,
    alifestd_to_working_format,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty():
    phylogeny_df = alifestd_make_empty()
    phylogeny_df["origin_time"] = None
    res = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
    )
    assert len(res) == 0


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple1(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Chain tree: 0 -> 1 -> 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_mark_clade_logistic_growth_children_asexual(
            phylogeny_df,
            mutate=mutate,
            parallel_backend=parallel_backend,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple2(mutate: bool, parallel_backend: typing.Optional[str]):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
            "origin_time": [50, 20, 10, 30, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[0, "clade_logistic_growth_children"] < 0
    assert result_df.loc[1, "clade_logistic_growth_children"] < 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple4(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
            "origin_time": [0, 10, 20, 40, 30, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "clade_logistic_growth_children"] > 0
    assert result_df.loc[1, "clade_logistic_growth_children"] < 0
    assert result_df.loc[2, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple5(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
            "origin_time": [0, 10, 200, 30, 40, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "clade_logistic_growth_children"] > 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[4, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple6(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Chain tree: 0
    phylogeny_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[None]"],
            "origin_time": [0],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[0, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple7(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
            "origin_time": [
                0,
                10,
                20,
                30,
                40,
                50,
                60,
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "clade_logistic_growth_children"] < 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[4, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple8(mutate: bool, parallel_backend: typing.Optional[str]):
    # Chain tree: 0
    phylogeny_df = pd.DataFrame(
        {
            "id": [1],
            "ancestor_list": ["[None]"],
            "origin_time": [0],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[1, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple9(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
            "origin_time": [0, 10, 5, 30, 40, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "clade_logistic_growth_children"] < 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[4, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple10(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
            "origin_time": [
                10,
                10,
                10,
                10,
                10,
                10,
                10,
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    _0 = pytest.approx(0, abs=1e-3)
    assert result_df.loc[0, "clade_logistic_growth_children"] == _0
    assert result_df.loc[1, "clade_logistic_growth_children"] == _0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[4, "clade_logistic_growth_children"] == _0
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple11(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0             7
    #       /   \          / \
    #      1     2        8   9
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
                "[None]",
                "[7]",
                "[7]",
            ],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_mark_clade_logistic_growth_children_asexual(
            phylogeny_df,
            mutate=mutate,
            parallel_backend=parallel_backend,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple12(mutate: bool, parallel_backend: typing.Optional[str]):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
            "origin_time": [50, 0, 20, 100, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[0, "clade_logistic_growth_children"] < 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple13(mutate: bool, parallel_backend: typing.Optional[str]):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 22
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 22, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
            "origin_time": [50, 0, 20, 10, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[0, "clade_logistic_growth_children"] < 0
    assert result_df.loc[1, "clade_logistic_growth_children"] < 0
    assert math.isnan(result_df.loc[22, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple14(mutate: bool, parallel_backend: typing.Optional[str]):
    # Tree structure:
    #        |---- 4
    #   |--- 100 --- 3
    #   1
    #   |--- 22
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 100, 1, 22, 4],
            "ancestor_list": ["[100]", "[1]", "[None]", "[1]", "[100]"],
            "origin_time": [50, 0, 20, 10, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert result_df.loc[100, "clade_logistic_growth_children"] < 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[22, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple15(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
            "origin_time": [0, 10, 20, 30, 40, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "clade_logistic_growth_children"] > 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert result_df.loc[2, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple16(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \   / \
    #    3   4 5   6
    #

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[2]",
            ],
            "origin_time": [0, 10, 20, 30, 40, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
        work_mask=np.array([False, True, False, False, True, False, False]),
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[0, "clade_logistic_growth_children"])
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("apply", [lambda x: x, alifestd_to_working_format])
@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple17(
    apply: typing.Callable,
    mutate: bool,
    parallel_backend: typing.Optional[str],
):
    # Tree structure:
    #         0
    #       /   \
    #      1     2
    #     / \
    #    3   4
    #       / \
    #      5   6

    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[0]",
                "[1]",
                "[1]",
                "[4]",
                "[4]",
            ],
            "origin_time": [0, 10, 200, 30, 40, 50, 60],
        },
    )
    phylogeny_df = apply(phylogeny_df)
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
        work_mask=[True, True, False, False, False, False, False],
    )
    result_df.index = result_df["id"]
    assert result_df.loc[0, "clade_logistic_growth_children"] > 0
    assert result_df.loc[1, "clade_logistic_growth_children"] > 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[5, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[6, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_simple18(mutate: bool, parallel_backend: typing.Optional[str]):
    # Tree structure:
    #        |---- 4
    #   |--- 0 --- 3
    #   1
    #   |--- 2
    phylogeny_df = pd.DataFrame(
        {
            "id": [3, 0, 1, 2, 4],
            "ancestor_list": ["[0]", "[1]", "[None]", "[1]", "[0]"],
            "origin_time": [50, 20, 10, 30, 40],
        },
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        mutate=mutate,
        parallel_backend=parallel_backend,
        work_mask=np.array([False, False, True, False, False]),
    )
    result_df.index = result_df["id"]
    assert math.isnan(result_df.loc[3, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[0, "clade_logistic_growth_children"])
    assert result_df.loc[1, "clade_logistic_growth_children"] < 0
    assert math.isnan(result_df.loc[2, "clade_logistic_growth_children"])
    assert math.isnan(result_df.loc[4, "clade_logistic_growth_children"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize("parallel_backend", [None, "loky"])
def test_fuzz(
    phylogeny_df: pd.DataFrame, parallel_backend: typing.Optional[str]
):
    if "origin_time" not in phylogeny_df.columns:
        phylogeny_df["origin_time"] = phylogeny_df["id"]

    phylogeny_df = alifestd_splay_polytomies(phylogeny_df)
    phylogeny_df = alifestd_collapse_unifurcations(phylogeny_df)
    phylogeny_df = alifestd_delete_unifurcating_roots_asexual(phylogeny_df)
    original = phylogeny_df.copy()

    result_df = alifestd_mark_clade_logistic_growth_children_asexual(
        phylogeny_df,
        parallel_backend=parallel_backend,
    )

    # Confirm that the input dataframe is not mutated.
    assert original.equals(phylogeny_df)

    # Verify result_df
    assert "clade_logistic_growth_children" in result_df.columns
    assert len(result_df) == len(phylogeny_df)

    result_df = alifestd_mark_leaves(result_df)
    assert (
        result_df["clade_logistic_growth_children"].isna()
        == result_df["is_leaf"]
    ).all()



import itertools as it
import warnings

import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit
from ._jit_numba_dict_t import jit_numba_dict_t
from ._jit_numpy_int64_t import jit_numpy_int64_t


@jit(nopython=True)
def _alifestd_splay_polytomies_fast_path(
    ancestor_ids: np.ndarray,
) -> pd.DataFrame:
    """Implementation detail for alifestd_splay_polytomies."""
    i64 = jit_numpy_int64_t
    splay_to_lookup = jit_numba_dict_t.empty(key_type=i64, value_type=i64)
    splay_from_lookup = jit_numba_dict_t.empty(key_type=i64, value_type=i64)

    new_rows = jit_numba_dict_t.empty(key_type=i64, value_type=i64)
    new_row_sources = jit_numba_dict_t.empty(key_type=i64, value_type=i64)
    for id_, ancestor_id in enumerate(ancestor_ids):
        if ancestor_id == id_:  # root
            continue

        if ancestor_id not in splay_to_lookup:
            splayed_from_id = ancestor_id
            splayed_to_id = ancestor_id
        else:
            splayed_from_id = splay_to_lookup[ancestor_id]
            splayed_to_id = len(ancestor_ids) + len(new_rows)

            ancestor_ids[id_] = splayed_to_id
            new_rows[splayed_to_id] = splayed_from_id
            new_row_sources[splayed_to_id] = ancestor_id

        splay_from_lookup[splayed_to_id] = splayed_from_id
        splay_to_lookup[ancestor_id] = splayed_to_id

    # drop unnecessary splays
    dropped_splay_remap = jit_numba_dict_t.empty(key_type=i64, value_type=i64)
    for ancestor_id, splayed_to_id in splay_to_lookup.items():
        if ancestor_id != splayed_to_id:
            splayed_from_id = splay_from_lookup[splayed_to_id]
            assert new_rows[splayed_to_id] == splayed_from_id
            dropped_splay_remap[splayed_to_id] = splayed_from_id
            del new_rows[splayed_to_id]
            del new_row_sources[splayed_to_id]

    # correct dropped splays
    for idx, ancestor_id in enumerate(ancestor_ids):
        splayed_to_id = dropped_splay_remap.get(ancestor_id, ancestor_id)
        ancestor_ids[idx] = splayed_to_id

    new_ids = np.array([*new_rows.keys()], dtype=np.int64)
    new_ancestor_ids = np.array([*new_rows.values()], dtype=np.int64)
    new_source_ids = np.array(
        [new_row_sources[new_id] for new_id in new_ids], dtype=np.int64
    )
    return ancestor_ids, new_source_ids, new_ids, new_ancestor_ids


def _alifestd_splay_polytomies_slow_path(
    phylogeny_df: pd.DataFrame,
) -> pd.DataFrame:
    """Implementation detail for alifestd_splay_polytomies."""
    splay_to_lookup = dict()
    splay_from_lookup = dict()
    phylogeny_df.set_index("id", drop=False, inplace=True)

    new_id_generator = it.count(phylogeny_df["id"].max() + 1)

    new_rows = dict()
    for id_, row in phylogeny_df.iterrows():
        ancestor_id = row["ancestor_id"]
        if ancestor_id == id_:  # root
            continue

        if ancestor_id not in splay_to_lookup:
            splayed_from_id = ancestor_id
            splayed_to_id = ancestor_id
        else:
            splayed_from_id = splay_to_lookup[ancestor_id]
            splayed_to_id = next(new_id_generator)

            phylogeny_df.at[id_, "ancestor_id"] = splayed_to_id
            new_row = phylogeny_df.loc[ancestor_id].copy()
            new_row["id"] = splayed_to_id
            new_row["ancestor_id"] = splayed_from_id
            new_rows[splayed_to_id] = new_row

        splay_from_lookup[splayed_to_id] = splayed_from_id
        splay_to_lookup[ancestor_id] = splayed_to_id

    # drop unnecessary splays
    dropped_splay_remap = dict()
    for ancestor_id, splayed_to_id in splay_to_lookup.items():
        if ancestor_id != splayed_to_id:
            splayed_from_id = splay_from_lookup[splayed_to_id]
            assert new_rows[splayed_to_id]["ancestor_id"] == splayed_from_id
            dropped_splay_remap[splayed_to_id] = splayed_from_id
            del new_rows[splayed_to_id]

    # correct dropped splays
    phylogeny_df["ancestor_id"] = (
        phylogeny_df["ancestor_id"]
        .map(
            {
                **{
                    ancestor_id: ancestor_id
                    for ancestor_id in phylogeny_df["ancestor_id"]
                },
                **dropped_splay_remap,
            }
        )
        .astype(int)
    )

    appendum_df = pd.DataFrame(new_rows.values())
    phylogeny_df = pd.concat([phylogeny_df, appendum_df], ignore_index=True)

    return phylogeny_df.reset_index(drop=True)


def alifestd_splay_polytomies(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    r"""Use a simple splay strategy to resolve polytomies, converting them into
    bifurcations.

    For example,
    ```
      1
     /|\
    2 3 4
    ```
    becomes
    ```
      1
     / \
    2   5
       / \
      3   4
    ```

    No adjustments to any branch length columns in `phylogeny_df` are
    performed. However, `origin_time` (as well as all other columns) of a
    polytomy's parent node are duplicated in splayed-out nodes that resolve
    that polytomy. So, nodes added to perform the splaying-out will have zero-
    length subtending branches in this regard (i.e., their origin time will
    match their parent's).

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if "branch_length" in phylogeny_df or "edge_length" in phylogeny_df:
        warnings.warn(
            "alifestd_splay_polytomies does not update branch length columns. "
            "Use `origin_time` to recalculate branch lengths for collapsed "
            "phylogeny."
        )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_asexual(phylogeny_df):
        raise NotImplementedError

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
        (
            phylogeny_df["ancestor_id"],
            new_source_ids,
            new_ids,
            new_ancestor_ids,
        ) = _alifestd_splay_polytomies_fast_path(
            phylogeny_df["ancestor_id"].to_numpy(dtype=np.int64)
        )
        addendum = phylogeny_df.loc[new_source_ids].copy()
        addendum["id"] = new_ids
        addendum["ancestor_id"] = new_ancestor_ids
        phylogeny_df = pd.concat([phylogeny_df, addendum], ignore_index=True)
    else:
        phylogeny_df = _alifestd_splay_polytomies_slow_path(phylogeny_df)

    if "ancestor_list" in phylogeny_df:
        phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
            phylogeny_df["id"], phylogeny_df["ancestor_id"]
        )

    return phylogeny_df
